import numpy as np
import copy
from typing import Optional

from tqdm import tqdm
from enum import Enum, auto, IntEnum
from io import StringIO
import os
import logging
import asyncio

from wrdice.D12Colored import D12Colored
from wrdice.util import *
from wrdice.Army import Army
from wrdice.Battle import Battle
from wrdice.config import *

def start_sim(in_, out_):
    asyncio.run(Simulator(in_, out_))

def Simulator(q_in, q_out):
    sim = Simulate(None, None)
    print('Starting Simulator Process')
    while True:
        #n = q_in.coro_get()
        n = q_in.get()
        if n is None:
            return
        else:
            # new simulation
            combat_system = n[0]
            config = n[1]
            army_a = n[2]
            army_b = n[3]
            sim.reset()
            msg = sim.run_all(combat_system, config, army_a, army_b)
        q_out.put(msg)

class Simulate:
    def __init__(self, army_a: Optional[Army], army_b: Optional[Army]):
        self.N = 500
        self.army_a = army_a
        self.army_b = army_b
        self.battle_type = 'land'

        self.statistics = []
        self.survivors = {'A':{'land':[],
                                'air':[],
                                'sea':[]},
                          'B':{'land':[],
                                'air':[],
                                'sea':[]}}

        self.metrics = {}


    def reset(self):
        self.battle_type = 'land'
        self.army_a = None
        self.army_b = None

        self.statistics = []
        self.survivors = {'A':{'land':[],
                                'air':[],
                                'sea':[]},
                          'B':{'land':[],
                                'air':[],
                                'sea':[]}}

        self.metrics = {}


    def update_battle_type(self):
        self.battle_type = 'land'
        if self.army_a is not None and self.army_b is not None:
            if np.sum(self.army_a.units['sea']) + np.sum(self.army_b.units['sea']) > 0:
                self.battle_type = 'sea'
        

    def run_all(self, combat_system: CombatSystem, config=None,  armyA: Optional[Army]=None, armyB: Optional[Army]=None) -> StringIO:
        if armyA is not None:
            self.army_a = armyA
        if armyB is not None:
            self.army_b = armyB

        self.update_battle_type()

        ret = self.run(combat_system, config)
        if ret:
            self.eval_statistics()
            msg = self.get_report()
            return msg
        else:
            return None

    def run(self, combat_system: CombatSystem, config=None) -> bool:
        troops_a, troops_b = 0, 0
        for T in ['sea', 'land', 'air']:
            troops_a += self.army_a.units[T].sum() 
            troops_b += self.army_b.units[T].sum() 

        if troops_a == 0 or troops_b == 0:
            return False

        if combat_system is CombatSystem.WarRoomV2 and config is None:
            logging.warning("NO CONFIG provided falling back to default config!!!!")
            config = wr20_vaniilla_options
        elif config is None:
            raise RuntimeError("No Config for combat system found")

        #logging.basicConfig(level=logging.DEBUG)
        #logging.debug(f"Army A Dice Ground {self.army_a.n_dice_ground}")
        #logging.debug(f"Army B Dice Ground {self.army_b.n_dice_ground}")
        #logging.debug(f"Army A Dice Air {self.army_a.n_dice_air}")
        #logging.debug(f"Army B Dice Air {self.army_b.n_dice_air}")


        for n in tqdm(range(self.N)):
            battle = Battle(copy.deepcopy(self.army_a), 
                            copy.deepcopy(self.army_b),
                            config)
            status = battle.run(combat_system=combat_system)
            self.statistics.append(status)

            for side in ['A', 'B']:
                for type in ['land', 'air', 'sea']:
                    self.survivors[side][type].append(battle.army[side].units[type])
                    if type == 'sea' and battle.army[side].submerged > 0:
                        self.survivors[side][type][-1][0] += battle.army[side].submerged
        return True


    def eval_statistics(self):
        # eval
        idx, count = np.unique(self.statistics, return_counts=True)
        stats = np.zeros(4)
        stats[idx] = count
        stats /= self.N
        self.statistics = np.array(self.statistics)

        self.survivors_a_ground = np.array(self.survivors['A'][self.battle_type])
        self.survivors_b_ground = np.array(self.survivors['B'][self.battle_type])

        self.survivors_a_air = np.array(self.survivors['A']['air'])
        self.survivors_b_air = np.array(self.survivors['B']['air'])

        self.survivors_a = np.concatenate((self.survivors_a_ground, self.survivors_a_air), 1)
        self.survivors_b = np.concatenate((self.survivors_b_ground, self.survivors_b_air), 1)

        idx_games_won_b = np.argwhere(self.statistics==0)
        idx_games_won_a = np.argwhere(self.statistics==1)
        idx_games_draw  = np.argwhere(self.statistics==2)
        idx_games_won_none = np.argwhere(self.statistics==3)

        avg_surv_a   = np.squeeze(np.mean(self.survivors_a[idx_games_won_a], 0))    if idx_games_won_a.any()    else np.array([0,0,0,0,0])
        avg_surv_b   = np.squeeze(np.mean(self.survivors_b[idx_games_won_b], 0))    if idx_games_won_b.any()    else np.array([0,0,0,0,0])
        avg_draw_a   = np.squeeze(np.mean(self.survivors_a[idx_games_draw], 0))     if idx_games_draw.any()     else np.array([0,0,0,0,0])
        avg_draw_b   = np.squeeze(np.mean(self.survivors_b[idx_games_draw], 0))     if idx_games_draw.any()     else np.array([0,0,0,0,0])

        self.metrics['avg_surv_a'] = avg_surv_a
        self.metrics['avg_surv_b'] = avg_surv_b
        self.metrics['avg_draw_a'] = avg_draw_a
        self.metrics['avg_draw_b'] = avg_draw_b

        games_won_a     = 0 if not idx_games_won_a.any()    else len(idx_games_won_a)
        games_won_b     = 0 if not idx_games_won_b.any()    else len(idx_games_won_b)
        games_draw      = 0 if not idx_games_draw.any()     else len(idx_games_draw)
        games_won_none  = 0 if not idx_games_won_none.any() else len(idx_games_won_none)

        self.metrics['games_won_a'] = games_won_a
        self.metrics['games_won_b'] = games_won_b
        self.metrics['games_won_none'] = games_won_none
        self.metrics['games_draw'] = games_draw


        def top_variations(games, N=10):
            variations, counts = np.unique(games, axis=0, return_counts=True)
            sorting = np.argsort(counts)
            N = min(N, len(variations))
            top_variations = variations[sorting][::-1][:N]
            percentages = (counts[sorting][::-1][:N] / len(games))

            if len(top_variations.shape) > 2:
                top_variations = np.squeeze(top_variations)
            if len(percentages.shape) > 2:
                percentages = np.squeeze(percentages)
            
            # for only one variation add count dimension
            if len(top_variations.shape) == 1:
                top_variations = top_variations[None, :]
            return top_variations.astype(int), percentages
        
        top_varations_won_a, distr_won_a = top_variations(self.survivors_a[idx_games_won_a])
        top_varations_won_b, distr_won_b = top_variations(self.survivors_b[idx_games_won_b])
        top_varations_draw_a, distr_draw_a = top_variations(self.survivors_a[idx_games_draw])
        top_varations_draw_b, distr_draw_b = top_variations(self.survivors_b[idx_games_draw])

        self.metrics['outcomes_won_a']  = {'variations': top_varations_won_a, 'distribution': distr_won_a}
        self.metrics['outcomes_won_b']  = {'variations': top_varations_won_b, 'distribution': distr_won_b}
        self.metrics['outcomes_draw_a'] = {'variations': top_varations_draw_a, 'distribution': distr_draw_a}
        self.metrics['outcomes_draw_b'] = {'variations': top_varations_draw_b, 'distribution': distr_draw_b}



    def get_report(self, till=COLOR.BLACK) -> str:
        report = StringIO(newline=os.linesep)

        games_won_a     = self.metrics['games_won_a']     
        games_won_b     = self.metrics['games_won_b']     
        games_won_none  = self.metrics['games_won_none']  
        games_draw      = self.metrics['games_draw']      

        
                                                          
        def print_avg_surv(data, id='') -> str:
            buf = StringIO()
            str = f"{id:<10}"
            for idx in range(till):
                str += f"{data[idx]:<10.2f}"
            # append air
            str += f"{data[5+COLOR.GREEN]:<10.2f}"
            str += f"{data[5+COLOR.RED]:<10.2f}"
            buf.write(str+os.linesep) 
            return buf.getvalue()
        
        def print_top_variations(data, N=3) -> str:
            buf = StringIO(newline=os.linesep)
            N = min(len(data['distribution']), N)

            for i in range(N):
                str = f"{data['distribution'][i]:<10.2f}"
                for color in COLOR:
                    if color == till:
                        break
                    str += f"{data['variations'][i][color.value]:<10}"
                #buf.write(f"{data['distribution'][i]:<10.2f} {(data['variations'][i][0]):<10} {(data['variations'][i][1]):<10} {(data['variations'][i][2]):<10} {(data['variations'][i][3]):<10} {(data['variations'][i][4]):<10}{os.linesep}") 
                # append air
                str += f"{data['variations'][i][5+COLOR.GREEN]:<10}"
                str += f"{data['variations'][i][5+COLOR.RED]:<10}"
                buf.write(str+os.linesep)
            return buf.getvalue()

        def print_legend() -> str:
            buf = StringIO()
            str = f"{'':<10}"
            for color in COLOR:
                if color == till:
                    break
                str += f"{color.name:<10}"
            # append air
            str += f"{'AIR Green':<10}"
            str += f"{'AIR Red':<10}"
            buf.write(str+os.linesep)
            return buf.getvalue()
       
        report.write(f"Results:{os.linesep}")
        report.write(f"{'A Won:':<20}{games_won_a / self.N:.2f}{os.linesep}")
        report.write(f"{'B Won:':<20}{games_won_b / self.N:.2f}{os.linesep}") 
        report.write(f"{'Draw:':<20}{games_draw / self.N:.2f}{os.linesep}")
        report.write(f"{'Mutual Annihilation:':<20}{games_won_none / self.N:.2f}{os.linesep}")

        if games_won_a > 0:
            report.write("++++++++++++++++++++++++++++++++++++++++++++"+os.linesep)
            report.write(f"Army A Won - {games_won_a/self.N:.2f}{os.linesep}")
            report.write(print_legend())
            report.write(print_avg_surv(self.metrics['avg_surv_a'], id='AVG'))
            report.write("--------------------------------------------------------------"+os.linesep)
            report.write(print_top_variations(self.metrics['outcomes_won_a']))

        if games_won_b > 0:
            report.write("++++++++++++++++++++++++++++++++++++++++++++"+os.linesep)
            report.write(f'Army B Won - {games_won_b/self.N:.2f}{os.linesep}')
            report.write(print_legend())
            report.write(print_avg_surv(self.metrics['avg_surv_b'], id='AVG'))
            report.write("--------------------------------------------------------------"+os.linesep)
            report.write(print_top_variations(self.metrics['outcomes_won_b']))

        if games_draw > 0:
            report.write("++++++++++++++++++++++++++++++++++++++++++++"+os.linesep)
            report.write(f'Draw - {games_draw/self.N:.2f}{os.linesep}')
            report.write(print_legend())
            report.write("--------------------------------------------------------------"+os.linesep)
            report.write(print_avg_surv(self.metrics['avg_draw_a'], id='A - avg'))
            report.write(print_top_variations(self.metrics['outcomes_draw_a']))
            report.write("--------------------------------------------------------------"+os.linesep)
            report.write(print_avg_surv(self.metrics['avg_draw_b'], id='B - avg'))
            report.write(print_top_variations(self.metrics['outcomes_draw_b']))

        return report.getvalue()


import numpy as np

from wrdice.util import COLOR

class D12Colored:
    def __init__(self, batchsize):
        ''' yellow, blue, green, red, black, white'''
        self.p = np.array([4/12, 3/12, 2/12, 1/12, 1/12, 1/12])
        self.P = np.cumsum(self.p)
        self.batchsize=batchsize

    def roll_once(self, r):
        if r >= self.P[4]:
            return COLOR.WHITE
        elif r >= self.P[3]:
            return COLOR.BLACK
        elif r >= self.P[2]:
            return COLOR.RED
        elif r >= self.P[1]:
            return COLOR.GREEN
        elif r >= self.P[0]:
            return COLOR.BLUE
        else:
            return COLOR.YELLOW
 
    def roll(self, n_dice):
        res = np.random.random_sample(n_dice)
        roll = [self.roll_once(r) for r in res]
        # sort rolls into bins
        bins = bins=np.arange(-0.5, len(self.p)+.5, 1)
        binned = np.histogram(roll, bins)[0]
        return binned




from wrdice.Army import Army
from wrdice.D12Colored import D12Colored
from wrdice.util import *

import numpy as np
import os
import logging

class Battle:
    def __init__(self, army_a: Army, army_b: Army, options):
        self.army = {'A' : army_a, 
                     'B' : army_b}
        self.options = options
        self.d12_batch = D12Colored(options['batch_size'])
        self.fa = ['A', 'B'] #both side have black and white
        self.battle_ground = None
        self.battle_air = None
        logging.basicConfig(level=logging.INFO)


    def get_num_colors_land(self, side):
        unique = np.unique(np.argwhere(self.army[side].units['land']))
        return len(unique)


    def get_force_advantage(self):
        num_colors_a = self.get_num_colors_land('A')
        num_colors_b = self.get_num_colors_land('B')
        if num_colors_a > num_colors_b:
            return ['A']
        elif num_colors_b > num_colors_a:
            return ['B']
        else:
            return ['A', 'B']


    def check_force_advantage(self, side):
        fa = self.get_force_advantage()
        if fa is None or side in fa:
            return True
        else:
            return False


    def expected_hits(self, source, target):
        P = self.d12_batch.p[:5]

        if source in self.fa:
            P = P + self.d12_batch.p[5]
        mu = (P * self.options['batch_size']) // self.options['hp']['land']
        exp = np.minimum(self.army[target].units['land'], mu)
        return exp


    def apply_color_selection(self, source, target, hits):
        if not self.options['color_select']:
            # return directly if we play without color selection
            return hits

        # we can only select as much enemy colors as target as we bring into the fight
        num_colors = self.get_num_colors_land(source)
        if self.options['color_select_strategy'] == ColorSelectionStrategy.MAX_NUM_HITS_FIRST:
            exp = self.expected_hits(source, target)
            # select the num_colors colors with the most units to hit realisticly 
            # as we later use the indexes to clear the hits we are interested in the rest
            color_select_to_delete = np.argsort(exp)[::-1][-num_colors:]

        elif self.options['color_select_strategy'] == ColorSelectionStrategy.MAX_MORALE_LOST_FIRST:
            exp = self.expected_hits(source, target)
            exp *= self.options['unit_morale_land']
            color_select_to_delete = np.argsort(exp)[::-1][:num_colors]

        elif self.options['color_select_strategy'] == ColorSelectionStrategy.MAX_COLOR_REDUCTION_FIRST:
            exp = self.expected_hits(source, target)
            exp_after = self.army[target].units['land'] - exp
            color_select = np.argwhere(np.logical_and(exp_after<=0, self.army[target].units['land']>0))
            color_select_to_delete = np.delete(np.arange(5), color_select)
            if len(color_select) == 0:
                # we cannot reduce one color directly... fallback to MAX_MORALE
                exp = self.expected_hits(source, target)
                exp *= self.options['unit_morale_land']
                color_select_to_delete = np.argsort(exp)[::-1][:num_colors]


        #apply color selection to hits_land
        hits[color_select_to_delete] = 0
        return hits



    def roll_air_and_apply_hits_wr2(self, source, target, batch):
        dice = self.options['batch_size'] if self.army[source].n_dice_air > self.options['batch_size'] else self.army[source].n_dice_air
        self.army[source].n_dice_air -= self.options['batch_size']
        self.army[source].n_dice_air = max(0, self.army[source].n_dice_air)

        roll = self.d12_batch.roll(dice)


        # get hp pools for planes
        hits_air = np.zeros(COLOR.BLACK)
        hits_air[:COLOR.BLACK] = roll[:COLOR.BLACK]


        air_hp = self.army[target].units_hp['air']
        priority = np.arange(len(air_hp.reshape(-1))).reshape(-1,2)

        for color in [COLOR.YELLOW, COLOR.BLUE, COLOR.GREEN, COLOR.RED]:
            for p in np.argsort(priority[color])[::-1]:
                while air_hp[color][p] > 0 and hits_air[color] > 0:
                    air_hp[color][p] -= 1
                    hits_air[color] -= 1


        if self.army[source].strategy['air'] == Strategy.BlackToHighestValueFirst:
            priority = np.argsort(self.options['unit_morale_air'])[::-1]
            p2 = priority*2+1
            priority = np.concatenate((priority, p2))

            for color in [COLOR.YELLOW, COLOR.BLUE, COLOR.GREEN, COLOR.RED]:
                for p in np.argsort(priority[color])[::-1]:
                    while air_hp[color][p] > 0 and hits_air[COLOR.BLACK] > 0:
                        air_hp[color][p] -= 1
                        hits_air[COLOR.BLACK] -= 1

            unit_hp = self.options['hp']['air']

            for color in [COLOR.YELLOW, COLOR.BLUE, COLOR.GREEN, COLOR.RED]:
                for p in np.argsort(priority[color])[::-1]:
                    while air_hp[color][p] > 0 and hits_air[COLOR.WHITE] > 0 and air_hp[color][p] % unit_hp[color][p]:
                        air_hp[color][p] -= 1
                        hits_air[COLOR.WHITE] -= 1

        self.army[target].units_hp['air'] = air_hp

    def roll_ground_and_apply_hits_wr2(self, source, target, batch):
        dice = self.options['batch_size'] if self.army[source].n_dice_ground > self.options['batch_size'] else self.army[source].n_dice_ground
        self.army[source].n_dice_ground -= self.options['batch_size']
        self.army[source].n_dice_ground = max(0, self.army[source].n_dice_ground)

        roll = self.d12_batch.roll(dice)

        # get hp pools for planes
        hits_ground = np.zeros(COLOR.BLACK)
        hits_ground[:COLOR.BLACK] = roll[:COLOR.BLACK]

        ground_hp = np.concatenate((self.army[target].units_hp['land'], 
                                    self.army[target].units_hp['sea']), axis=1)
        unit_hp = np.concatenate((self.options['hp']['land'], 
                                   self.options['hp']['sea'])).T
        unit_values = np.concatenate((self.options['unit_attack']['land'][STANCE.GROUND], 
                                      self.options['unit_attack']['sea'][STANCE.GROUND])).T

        # kill of units with less hp first -> more dead
        priority = np.argsort(unit_hp)
        # we have 4 priorities for each color now
        for color in [COLOR.YELLOW, COLOR.BLUE, COLOR.GREEN, COLOR.RED]:
            for p in priority[color][::-1]:
                while ground_hp[color][p] > 0 and hits_ground[color] > 0:
                    # check for escort only use to prevent sinking elsewise let the big ship soak
                    # a bit of damage first
                    if self.battle_ground == 'sea' and\
                            color in [COLOR.GREEN, COLOR.RED] and\
                            ground_hp[color][p] % self.options['hp']['sea'][0][color] == 1 and\
                            np.any(ground_hp[COLOR.BLUE][2] > 0):
                        ground_hp[COLOR.BLUE][2] -= 1
                    else:
                        # apply damage regularily
                        ground_hp[color][p] -= 1

                    hits_ground[color] -= 1

        
        if (source in self.fa and \
            self.army[source].strategy['ground'] == Strategy.BlackToHighestValueFirst):
            priority = np.argsort(unit_values)
            #BLACK
            for color in [COLOR.YELLOW, COLOR.BLUE, COLOR.GREEN, COLOR.RED]:
                for p in priority[color][::-1]:
                    while ground_hp[color][p] > 0 and hits_ground[COLOR.BLACK] > 0:
                        ground_hp[color][p] -= 1
                        hits_ground[COLOR.BLACK] -= 1

            #WHITE
            for color in [COLOR.YELLOW, COLOR.BLUE, COLOR.GREEN, COLOR.RED]:
                for p in np.argsort(priority[color])[::-1]:
                    while ground_hp[color][p] > 0 and hits_ground[COLOR.WHITE] > 0 and ground_hp[color][p] % unit_hp[p]:
                        ground_hp[color][p] -= 1
                        hits_ground[COLOR.WHITE] -= 1

        # write back updated hp pool
        land_hp, sea_hp = np.hsplit(ground_hp, 2)
        self.army[target].units_hp['land'] = land_hp
        self.army[target].units_hp['sea'] = sea_hp




        
    def update_unit_count(self, target):
        # update unit count
        for type in ['air', 'land', 'sea']:
            with np.errstate(divide='ignore'):
                div = np.divide(self.army[target].units_hp[type].T, self.options['hp'][type])
            # fix divide by 0 hp
            div[np.isnan(div)] = 0
            in_battle = np.ceil(div)

            # subs flee battle
            # TODO: add flag/config to indicate submersibles
            if type == 'sea' and np.any(self.army[target].units_hp['sea'][0] % self.options['hp']['sea'][:,0] == 1):
                idx = np.argwhere(self.army[target].units_hp['sea'][0] % self.options['hp']['sea'][:,0] == 1)
                self.army[target].submerged += 1
                self.army[target].units_hp['sea'][0][idx] -= 1
                self.army[target].units_by_stance['sea'][0][idx] -= 1

            in_battle_combined = np.sum(in_battle, axis=0) # sum over different stances
            self.army[target].units[type] = in_battle_combined
            self.army[target].units_by_stance[type] = in_battle




    def roll_air_and_apply_hits_wr2qb(self, source, target):
        dice = self.options['batch_size'] if self.army[source].n_dice_air > self.options['batch_size'] else self.army[source].n_dice_air
        self.army[source].n_dice_air -= self.options['batch_size']
        self.army[source].n_dice_air = max(0, self.army[source].n_dice_air)

        roll = self.d12_batch.roll(dice)
        # get hp pools for planes
        hp_air_stance_0 = self.ptions['hp_air'][0]
        hp_air_stance_1 = self.options['hp_air'][1]
        
        hits_air = np.zeros(5)
        unmatched_air = np.zeros(5)

        hits_air[:COLOR.BLACK] = roll[:COLOR.BLACK] // hp_air_stance_0[:COLOR.BLACK]
        unmatched_air[:COLOR.BLACK] = roll[:COLOR.BLACK] % hp_air_stance_0[COLOR.BLACK]
        # reset black
        unmatched_air[COLOR.BLACK] += hits_air[COLOR.BLACK]
        hits_air[COLOR.BLACK] = 0
        # apply direct hits to all but black
        self.army[target].units['air'][:COLOR.BLACK] = np.maximum(self.army[target].units['air'][:COLOR.BLACK] - hits_air[:COLOR.BLACK], [0, 0, 0, 0, 0])
        wild = unmatched_air[COLOR.BLACK]
        while wild > 0:
            if wild == 0:
                break # we matched all blacks
            hits = unmatched_air > 0
            trgt = self.army[target].units['land'] > 0
            unit_targets = np.logical_and(hits[:COLOR.BLACK], trgt)

            # find target high to low
            for color in range(COLOR.RED,-1,-1):
                if unit_targets[color] and wild > 0:
                    wild -= 1
                    self.army[target].units['land'][color] -= 1
                    #found one
                    break
            #none found
            wild = 0







    def roll_land_and_apply_hits(self, source, target, batch):

        dice = self.options['batch_size'] if self.army[source].n_dice_ground > self.options['batch_size'] else self.army[source].n_dice_ground
        self.army[source].n_dice_ground -= self.options['batch_size']
        self.army[source].n_dice_ground = max(0, self.army[source].n_dice_ground)

        roll = self.d12_batch.roll(dice)


        hits_land = roll // self.options['hp_land']
        unmatched_land = roll % self.options['hp_land']
        # reset black
        unmatched_land[5] += hits_land[5]
        hits_land[5] = 0

        hits_land = self.apply_color_selection(source, target, hits_land)

        # apply direct hits to all but black
        self.army[target].units['land'][:5] = np.maximum(self.army[target].units['land'][:5] - hits_land[:5], [0, 0, 0, 0, 0])

        if source in self.fa or not self.options['force_advantage']:
            # apply remainder to highest valued unit
            wild = unmatched_land[5]
            while wild > 0:
                if wild == 0:
                    break # we matched all blacks
                hits = unmatched_land > 0
                trgt = self.army[target].units['land'] > 0
                unit_targets = np.logical_and(hits[:5], trgt)

                # find target high to low
                for type in range(4,-1,-1):
                    if unit_targets[type] and wild > 0:
                        wild -= 1
                        self.army[target].units['land'][type] -= 1
                        #found one
                        break
                #none found
                wild = 0


    def check_batch_cap(self):
        # count colors (warroom 2.0 rules)
        count_a = (self.army['A'].units['land'] > 0).sum()
        count_b = (self.army['B'].units['land'] > 0).sum()
        
        if count_a > count_b:
            diff = 3-(count_a - count_b)
            cap = diff * self.options['batch_size']
            self.army['B'].n_dice_ground = min(self.army['B'].n_dice_ground, cap)
            logging.debug(f"Batch Cap - Limiting Army B to {self.army['B'].n_dice_ground} dice")
        elif count_b > count_a:
            diff = 3-(count_b - count_a)
            cap = diff * self.options['batch_size']
            self.army['A'].n_dice_ground = min(self.army['A'].n_dice_ground, cap)
            logging.debug(f"Batch Cap - Limiting Army A to {self.army['A'].n_dice_ground} dice")

    def run_warroomv2(self):
        if self.options['force_advantage']:
            self.fa = self.get_force_advantage()

        for batch in range(3):
            self.roll_air_and_apply_hits_wr2('A', 'B', batch)
            self.roll_air_and_apply_hits_wr2('B', 'A', batch)

            self.update_unit_count('A')
            self.update_unit_count('B')
        
        # update land combat strenght
        self.army['A'].update_dice_ground()
        self.army['B'].update_dice_ground()

        if self.options['batch_cap']:
            self.check_batch_cap()

        for batch in range(3):
            self.roll_ground_and_apply_hits_wr2('A', 'B', batch)
            self.roll_ground_and_apply_hits_wr2('B', 'A', batch)

            self.update_unit_count('A')
            self.update_unit_count('B')

            if self.options['recheck_force_advantage']:
                self.fa = self.get_force_advantage()
        
    def _print_armies(self):
        for type in ['land', 'sea', 'air']:
            print(type)
            for A in ['A', 'B']:
                print(f"{A:<10} {self.army[A].units[type]}")

    def run(self, combat_system: CombatSystem = CombatSystem.WarRoomV2) -> int:
        '''
            returncode
            0 - A lost
            1 - B lost
            2 - draw both survived
            3 - draw both eliminated
        '''

        # auto determine battle type by units present
        type = ['land', 'sea'] 
        self.battle_ground = 'land'
        for A in ['A','B']:
            for T in type:
                if self.army[A].units[T].sum() > 0:
                    self.battle_ground = T

        if combat_system == CombatSystem.WarRoomV2:
            self.run_warroomv2()
        elif combat_system == CombatSystem.WarRoomV2Quickbattle:
            self.run_warroomv2_quickbattle()
        elif combat_system == CombatSystem.AreWeTheBaddies:
            self.run_arewethebaddies()
        

        status = None
        if (self.army['A'].units[self.battle_ground].sum() == 0  and \
                (self.army['B'].units[self.battle_ground].sum() > 0 or \
                self.army['B'].units['air'].sum() > 0 )):
            status = 0

        elif ((self.army['A'].units[self.battle_ground].sum() > 0  or \
                self.army['A'].units['air'].sum() > 0)  and \
                self.army['B'].units[self.battle_ground].sum() == 0):
            status = 1

        elif (self.army['A'].units[self.battle_ground].sum() > 0  and \
                self.army['B'].units[self.battle_ground].sum() > 0):
            status = 2

        elif (self.army['A'].units[self.battle_ground].sum() == 0 and \
                self.army['B'].units[self.battle_ground].sum() == 0):
            status = 3
        
        return status




import math
import numpy as np
import copy
import pdb
from tqdm import tqdm
from enum import Enum, auto

class ColorSelectionStrategy(Enum):
    MAX_NUM_HITS_FIRST = auto()
    MAX_MORALE_LOST_FIRST = auto()
    MAX_COLOR_REDUCTION_FIRST = auto() 


'''
cap_total_dice      -   upper limit for dice rolled in one battle
force_advantage     -   black and white are not considered when fighting against more color variety on land
'''
options = {

    'cap_total_dice' : True,
    'cap_total_dice_limit': 30,

    'batch_cap':    True,
    'batch_size':   10,
    'color_select': True,

    'force_advantage': True,     
    'recheck_force_advantage': True,

    'color_select_strategy': ColorSelectionStrategy.MAX_COLOR_REDUCTION_FIRST,

    'hp_land': 2,
    'hp_air':  3,
    'hp_sea':  3,

    'unit_attack_land' : [1, 2, 3, 4, 6],
    'unit_attack_sea'  : [1, 2, 1, 1, 0],
    'unit_attack_air'  : [2, 1, 4, 6, 0],

    'unit_morale_land' : [1, 2, 4, 8, 12],
    'unit_morale_sea'  : [4, 8, 10, 12, 0],
    'unit_morale_air'  : [2, 4, 6, 10, 0]

}

class d12colored:
    def __init__(self, batchsize):
        ''' yellow, blue, green, red, white, black'''
        self.p = np.array([4/12, 3/12, 2/12, 1/12, 1/12, 1/12])
        self.P = np.cumsum(self.p)
        self.batchsize=batchsize

    def __r(self, r):
        if r >= self.P[4]:
            return 5
        elif r >= self.P[3]:
            return 4
        elif r >= self.P[2]:
            return 3
        elif r >= self.P[1]:
            return 2
        elif r >= self.P[0]:
            return 1
        else:
            return 0
 
    def roll(self, n_dice):
        res = np.random.random_sample(n_dice)
        roll = [self.__r(r) for r in res]
        # sort rolls into bins
        bins = bins=np.arange(-0.5, len(self.p)+.5, 1)
        binned = np.histogram(roll, bins)[0]
        return binned



class Army:
    def __init__(self, units_land, units_air, units_sea):
        self.units_land = units_land
        self.units_sea  = units_sea
        self.units_air  = units_air

        self.n_dice_ground = np.sum(units_land * options['unit_attack_land'] + \
                              units_air * options['unit_attack_air'] + \
                              units_sea * options['unit_attack_sea'])

        self.n_dice_air = np.sum(units_land * options['unit_attack_land'] + \
                              units_air * options['unit_attack_air'] + \
                              units_sea * options['unit_attack_sea'])



        if options['cap_total_dice']:
            self.n_dice_ground = np.minimum(self.n_dice_ground, options['cap_total_dice_limit'])
            self.n_dice_air    = np.minimum(self.n_dice_air,    options['cap_total_dice_limit'])

        


class Battle:
    def __init__(self, army_a: Army, army_b: Army):
        self.army = {'A' : army_a, 
                     'B' : army_b}

        self.d12_batch = d12colored(options['batch_size'])

        self.fa = ['A', 'B'] #both side have black and white

    def get_num_colors_land(self, side):
        unique = np.unique(np.argwhere(self.army[side].units_land))
        return len(unique)


    def get_force_advantage(self):
        num_colors_a = self.get_num_colors_land('A')
        num_colors_b = self.get_num_colors_land('B')
        if num_colors_a > num_colors_b:
            return ['A']
        elif num_colors_b > num_colors_a:
            return ['B']
        else:
            return ['A', 'B']


    def check_force_advantage(self, side):
        fa = self.get_force_advantage()
        if fa is None or side in fa:
            return True
        else:
            return False


    def expected_hits(self, source, target):
        P = self.d12_batch.p[:5]

        if source in self.fa:
            P = P + self.d12_batch.p[5]

        mu = (P * options['batch_size']) // options['hp_land']
        exp = np.minimum(self.army[target].units_land, mu)
        return exp

    def roll_air_and_apply_hits(self, source, target):
        dice = options['batch_size'] if self.army[source].n_dice_air > options['batch_size'] else self.army[source].n_dice_air
        self.army[source].n_dice_air -= options['batch_size']
        self.army[source].n_dice_air = max(0, self.army[source].n_dice_air)

        roll = self.d12_batch.roll(dice)

        hits_air = roll // options['hp_air']
        unmatched_air = roll % options['hp_air']
        # reset black
        unmatched_air[5] += hits_air[5]
        hits_air[5] = 0
        # apply direct hits to all but black
        self.army[target].units_air[:5] = np.maximum(self.army[target].units_air[:5] - hits_air[:5], [0, 0, 0, 0, 0])
        wild = unmatched_air[5]
        while wild > 0:
            if wild == 0:
                break # we matched all blacks
            hits = unmatched_air > 0
            trgt = self.army[target].units_land > 0
            unit_targets = np.logical_and(hits[:5], trgt)

            # find target high to low
            for type in range(4,-1,-1):
                if unit_targets[type] and wild > 0:
                    wild -= 1
                    self.army[target].units_land[type] -= 1
                    #found one
                    break
            #none found
            wild = 0





    def apply_color_selection(self, source, target, hits):
        if not options['color_select']:
            # return directly if we play without color selection
            return hits

        # we can only select as much enemy colors as target as we bring into the fight
        num_colors = self.get_num_colors_land(source)
        if options['color_select_strategy'] == ColorSelectionStrategy.MAX_NUM_HITS_FIRST:
            exp = self.expected_hits(source, target)
            # select the num_colors colors with the most units to hit realisticly 
            # as we later use the indexes to clear the hits we are interested in the rest
            color_select_to_delete = np.argsort(exp)[::-1][-num_colors:]

        elif options['color_select_strategy'] == ColorSelectionStrategy.MAX_MORALE_LOST_FIRST:
            exp = self.expected_hits(source, target)
            exp *= options['unit_morale_land']
            color_select_to_delete = np.argsort(exp)[::-1][:num_colors]

        elif options['color_select_strategy'] == ColorSelectionStrategy.MAX_COLOR_REDUCTION_FIRST:
            exp = self.expected_hits(source, target)
            exp_after = self.army[target].units_land - exp
            color_select = np.argwhere(np.logical_and(exp_after<=0, self.army[target].units_land>0))
            color_select_to_delete = np.delete(np.arange(5), color_select)
            if len(color_select) == 0:
                # we cannot reduce one color directly... fallback to MAX_MORALE
                exp = self.expected_hits(source, target)
                exp *= options['unit_morale_land']
                color_select_to_delete = np.argsort(exp)[::-1][:num_colors]


        #apply color selection to hits_land
        hits[color_select_to_delete] = 0
        return hits



    def roll_land_and_apply_hits(self, source, target):

        dice = options['batch_size'] if self.army[source].n_dice_ground > options['batch_size'] else self.army[source].n_dice_ground
        self.army[source].n_dice_ground -= options['batch_size']
        self.army[source].n_dice_ground = max(0, self.army[source].n_dice_ground)

        roll = self.d12_batch.roll(dice)


        hits_land = roll // options['hp_land']
        unmatched_land = roll % options['hp_land']
        # reset black
        unmatched_land[5] += hits_land[5]
        hits_land[5] = 0

        hits_land = self.apply_color_selection(source, target, hits_land)

        # apply direct hits to all but black
        self.army[target].units_land[:5] = np.maximum(self.army[target].units_land[:5] - hits_land[:5], [0, 0, 0, 0, 0])

        if source in self.fa or not options['force_advantage']:
            # apply remainder to highest valued unit
            wild = unmatched_land[5]
            while wild > 0:
                if wild == 0:
                    break # we matched all blacks
                hits = unmatched_land > 0
                trgt = self.army[target].units_land > 0
                unit_targets = np.logical_and(hits[:5], trgt)

                # find target high to low
                for type in range(4,-1,-1):
                    if unit_targets[type] and wild > 0:
                        wild -= 1
                        self.army[target].units_land[type] -= 1
                        #found one
                        break
                #none found
                wild = 0


    def check_batch_cap(self):
        # count colors (warroom 2.0 rules)
        count_a = np.argwhere(self.army['A'].units_land).sum()
        count_b = np.argwhere(self.army['B'].units_land).sum()
        if count_a > count_b:
            diff = count_a - count_b
            dice_diff = diff * options['batch_size']
            self.army['B'].n_dice_ground -= dice_diff
            #give em at least one batch
            self.army['B'].n_dice_ground = max(self.army['B'].n_dice_ground, options['batch_size'])
        elif count_b > count_a:
            diff = count_b - count_a
            dice_diff = diff * options['batch_size']
            self.army['A'].n_dice_ground -= dice_diff
            #give em at least one batch
            self.army['A'].n_dice_ground = max(self.army['A'].n_dice_ground, options['batch_size'])


    def run(self):
        if options['batch_cap']:
            self.check_batch_cap()

        if options['force_advantage']:
            self.fa = self.get_force_advantage()

        for batch in range(3):
            self.roll_air_and_apply_hits('A', 'B')
            self.roll_air_and_apply_hits('B', 'A')

        for batch in range(3):
            self.roll_land_and_apply_hits('A', 'B')
            self.roll_land_and_apply_hits('B', 'A')

            if options['recheck_force_advantage']:
                self.fa = self.get_force_advantage()

        '''
            returncode
            0 - A lost
            1 - B lost
            2 - draw both survived
            3 - draw both eliminated
        '''
        
        status = None
        if self.army['A'].units_land.sum() == 0 and self.army['B'].units_land.sum() > 0:
            status = 0
        elif self.army['A'].units_land.sum() > 0 and self.army['B'].units_land.sum() == 0:
            status = 1
        elif self.army['A'].units_land.sum() > 0 and self.army['B'].units_land.sum() > 0:
            status = 2
        elif self.army['A'].units_land.sum() == 0 and self.army['B'].units_land.sum() == 0:
            status = 3

        return (status, self.army['A'].units_land, self.army['B'].units_land)


class Simulate:
    def __init__(self, army_a: Army, army_b: Army):
        self.N = 5000
        self.army_a = army_a
        self.army_b = army_b
        self.statistics = []
        self.survivors_a = []
        self.survivors_b = []

        self.metrics = {}

    def run(self):
        for n in tqdm(range(self.N)):
            result = Battle(copy.deepcopy(self.army_a), copy.deepcopy(self.army_b)).run()
            self.statistics.append(result[0])
            self.survivors_a.append(result[1])
            self.survivors_b.append(result[2])

        self.eval_statistics()
        self.print_results()



    def eval_statistics(self):
        # eval
        idx, count = np.unique(self.statistics, return_counts=True)
        stats = np.zeros(4)
        stats[idx] = count
        stats /= self.N
        self.statistics = np.array(self.statistics)
        self.survivors_a = np.array(self.survivors_a)
        self.survivors_b = np.array(self.survivors_b)

        idx_games_won_b = np.argwhere(self.statistics==0)
        idx_games_won_a = np.argwhere(self.statistics==1)
        idx_games_draw  = np.argwhere(self.statistics==2)
        idx_games_won_none = np.argwhere(self.statistics==3)

        avg_surv_a   = np.squeeze(np.mean(self.survivors_a[idx_games_won_a], 0))    if idx_games_won_a.any()    else np.array([0,0,0,0,0])
        avg_surv_b   = np.squeeze(np.mean(self.survivors_b[idx_games_won_b], 0))    if idx_games_won_b.any()    else np.array([0,0,0,0,0])
        avg_draw_a   = np.squeeze(np.mean(self.survivors_a[idx_games_draw], 0))     if idx_games_draw.any()     else np.array([0,0,0,0,0])
        avg_draw_b   = np.squeeze(np.mean(self.survivors_b[idx_games_draw], 0))     if idx_games_draw.any()     else np.array([0,0,0,0,0])

        self.metrics['avg_surv_a'] = avg_surv_a
        self.metrics['avg_surv_b'] = avg_surv_b
        self.metrics['avg_draw_a'] = avg_draw_a
        self.metrics['avg_draw_b'] = avg_draw_b

        games_won_a     = 0 if not idx_games_won_a.any()    else len(idx_games_won_a)
        games_won_b     = 0 if not idx_games_won_b.any()    else len(idx_games_won_b)
        games_draw      = 0 if not idx_games_draw.any()     else len(idx_games_draw)
        games_won_none  = 0 if not idx_games_won_none.any() else len(idx_games_won_none)

        self.metrics['games_won_a'] = games_won_a
        self.metrics['games_won_b'] = games_won_b
        self.metrics['games_won_none'] = games_won_none
        self.metrics['games_draw'] = games_draw
        
        def top_variations(games, N=10):
            variations, counts = np.unique(games, axis=0, return_counts=True)
            sorting = np.argsort(counts)
            N = min(N, len(variations))
            top_variations = np.squeeze(variations[sorting][::-1][:N])
            percentages = np.squeeze(counts[sorting][::-1][:N] / len(games))
            return top_variations, percentages
        
        top_varations_won_a, distr_won_a = top_variations(self.survivors_a[idx_games_won_a])
        top_varations_won_b, distr_won_b = top_variations(self.survivors_a[idx_games_won_b])
        top_varations_draw_a, distr_draw_a = top_variations(self.survivors_a[idx_games_draw])
        top_varations_draw_b, distr_draw_b = top_variations(self.survivors_b[idx_games_draw])

        self.metrics['outcomes_won_a']  = {'variations': top_varations_won_a, 'distribution': distr_won_a}
        self.metrics['outcomes_won_b']  = {'variations': top_varations_won_b, 'distribution': distr_won_b}
        self.metrics['outcomes_draw_a'] = {'variations': top_varations_draw_a, 'distribution': distr_draw_a}
        self.metrics['outcomes_draw_b'] = {'variations': top_varations_draw_b, 'distribution': distr_draw_b}



    def print_results(self):
        games_won_a     = self.metrics['games_won_a']     
        games_won_b     = self.metrics['games_won_b']     
        games_won_none  = self.metrics['games_won_none']  
        games_draw      = self.metrics['games_draw']      
                                                    
        def print_avg_surv(data, id=''):
            print(f"{id:<10} {(data[0]):<10.2f} {(data[1]):<10.2f} {(data[2]):<10.2f} {(data[3]):<10.2f} {(data[4]):<10.2f}") 

        def print_top_variations(data, N=3):
            N = min(len(data['distribution']), N)

            for i in range(N):
                print(f"{data['distribution'][i]:<10.2f} {(data['variations'][i][0]):<10} {(data['variations'][i][1]):<10} {(data['variations'][i][2]):<10} {(data['variations'][i][3]):<10} {(data['variations'][i][4]):<10}") 


       
        print("Results:")
        print()
        print(f"{'A Lost:':<10}{games_won_b:<10}({games_won_b / self.N:.2f})") 
        print(f"{'B Lost':<10}{games_won_a:<10}({games_won_a / self.N:.2f})")
        print(f"{'both surv':<10}{games_draw:<10}({games_draw / self.N:.2f})")
        print(f"{'both died':<10}{games_won_none:<10}({games_won_none / self.N:.2f})")

        if games_won_a > 0:
            print()
            print("Army A Won")
            print(f"{'':<10} {'Yellow':<10} {'Blue':<10} {'Green':<10} {'Red':<10} {'White':<10}") 
            print_avg_surv(self.metrics['avg_surv_a'])
            print()
            print('average survivors A')
            print_avg_surv(self.metrics['avg_surv_a'])
            print()
            print('average survivors A')
            print("Top Variations")
            print_top_variations(self.metrics['outcomes_won_a'])

        if games_won_b > 0:
            print()
            print('Army B Won')
            print(f"{'':<10} {'Yellow':<10} {'Blue':<10} {'Green':<10} {'Red':<10} {'White':<10}") 
            print_avg_surv(self.metrics['avg_surv_b'])
            print()
            print('average survivors B')
            print_avg_surv(self.metrics['avg_surv_b'])
            print()
            print_top_variations(self.metrics['outcomes_won_b'])

        if games_draw > 0:
            print()
            print('Draw ')
            print(f"{'':<10} {'Yellow':<10} {'Blue':<10} {'Green':<10} {'Red':<10} {'White':<10}") 
            print_avg_surv(self.metrics['avg_draw_a'], id='A - avg')
            print_top_variations(self.metrics['outcomes_draw_a'])
            print()
            print_avg_surv(self.metrics['avg_draw_b'], id='B - avg')
            print_top_variations(self.metrics['outcomes_draw_b'])



if __name__ == '__main__':
    A = Army(np.array([3, 1, 3, 1, 0]), np.array([0, 0, 0, 0, 0]), np.array([0, 0, 0, 0, 0]))
    B = Army(np.array([0, 0, 0, 8, 0]), np.array([0, 0, 0, 0, 0]), np.array([0, 0, 0, 0, 0]))
    Simulate(A, B).run()
